package com.bergerkiller.bukkit.nolagg;

import java.io.DataInput;
import java.io.DataInputStream;
import java.io.DataOutput;
import java.io.DataOutputStream;
import java.io.File;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.zip.ZipException;

import org.bukkit.Bukkit;
import org.bukkit.craftbukkit.CraftWorld;

import net.minecraft.server.*;

@SuppressWarnings("rawtypes")
public class BufferedChunkRegionLoader extends ChunkRegionLoader implements IChunkLoader {
	public static int saved = 0;
		
	public static void swapAll() {
		if (!NoLagg.useBufferedChunkLoading) return;
		for (org.bukkit.World world : Bukkit.getServer().getWorlds()) {
			swap(world);
		}
	}
	public static void swap(org.bukkit.World world) {
		swap(((CraftWorld) world).getHandle());
	}
	public static void swap(WorldServer world) {
		swap(world.chunkProviderServer);
	}
	public static void swap(ChunkProviderServer cps) {
		if (!NoLagg.useBufferedChunkLoading) return;
		try {
			if (eraseChunkMethod == null) {
				eraseChunkMethod = RegionFile.class.getDeclaredMethod("a", int.class, int.class, int.class);
				eraseChunkMethod.setAccessible(true);
			}
			
			Field e = ChunkProviderServer.class.getDeclaredField("e");
			e.setAccessible(true);
			IChunkLoader loader = (IChunkLoader) e.get(cps);
			//get needed fields
			Field a = ChunkRegionLoader.class.getDeclaredField("a");
			Field b = ChunkRegionLoader.class.getDeclaredField("b");
			Field c = ChunkRegionLoader.class.getDeclaredField("c");
			Field d = ChunkRegionLoader.class.getDeclaredField("d");
			a.setAccessible(true);
			b.setAccessible(true);
			c.setAccessible(true);
			d.setAccessible(true);
			
			//get old loader variables
			List vara = (List) a.get(loader);
			Set varb = (Set) b.get(loader);
			Object varc = (Object) c.get(loader);
			File vard = (File) d.get(loader);
			
			//clear old loader's data
			a.set(loader, new ArrayList());
			b.set(loader, new HashSet());
			c.set(loader, new Object());
			
			//create a new loader and set it's native members
			loader = new BufferedChunkRegionLoader(cps.world, vara, varb, varc, vard);
			a.set(loader, vara);
			b.set(loader, varb);
			c.set(loader, varc);
			d.set(loader, vard);
			

			
			e.set(cps, loader);
		} catch (Throwable t) {
			NoLagg.log(Level.SEVERE, "Failed to initialize the buffered chunk loader:");
			t.printStackTrace();
			NoLagg.useBufferedChunkLoading = false;
		}
	}
	
	private static Method eraseChunkMethod;

	private final WorldServer worldOwner;
	private final List savequeue;
	private final Set pendingsave;
	private final Object lock;
	private final File sourcefolder;

	private BufferedChunkRegionLoader(final WorldServer world, List savequeue, Set pendingsave, Object lock, File sourcefolder) {
		super(sourcefolder);
		this.savequeue = savequeue;
		this.pendingsave = pendingsave;
		this.lock = lock;
		this.sourcefolder = sourcefolder;
		this.worldOwner = world;
	}

	private static Field pendinga;
	private static Field pendingb;
	public static ChunkCoordIntPair getPendingA(Object from) {
		try {
			if (pendinga == null) {
				pendinga = from.getClass().getDeclaredField("a");
				pendinga.setAccessible(true);
			}
			return (ChunkCoordIntPair) pendinga.get(from);
		} catch (Throwable t) {
			t.printStackTrace();
			return null;
		}
	}
	public static NBTTagCompound getPendingB(Object from) {
		try {
			if (pendingb == null) {
				pendingb = from.getClass().getDeclaredField("b");
				pendingb.setAccessible(true);
			}
			return (NBTTagCompound) pendingb.get(from);
		} catch (Throwable t) {
			t.printStackTrace();
			return null;
		}
	}

	@Override
    public boolean c() {
        Object pendingchunktosave = null;
        ChunkCoordIntPair coord;
        synchronized (this.lock) {
            if (this.savequeue.size() <= 0) {
                return false;
            }

            pendingchunktosave = this.savequeue.remove(0);
            coord = getPendingA(pendingchunktosave);
            this.pendingsave.remove(coord);
        }
        
        if (pendingchunktosave != null) {
            try {
                DataOutputStream dataoutputstream = RegionFileCache.c(this.sourcefolder, coord.x, coord.z);
                NBTCompressedStreamTools.a(getPendingB(pendingchunktosave), (DataOutput) dataoutputstream);
                saved++;
                dataoutputstream.flush();
                dataoutputstream.close();
                BufferedChunk chunk = BufferedChunk.removeUnloaded(this.worldOwner, coord.x, coord.z);
                if (chunk != null && chunk.isUnloaded()) {
        			chunk.setFree();
                }
            } catch (Exception exception) {
                exception.printStackTrace();
            }
        }
        return true;
    }
		
	public void clearChunk(int x, int z) throws Throwable {
		RegionFile file = RegionFileCache.a(this.sourcefolder, x, z);
		try {
			eraseChunkMethod.invoke(file, x & 31, z & 31, 0);
		} catch (InvocationTargetException ex) {
			throw ex.getCause();
		}
	}
	
	/*
	 * Chunk at wrong location: relocating code
	 */
	public void relocate(BufferedChunk chunk, NBTTagCompound compound, int realX, int realZ) throws Throwable {
		//is a chunk present at the old location?
		NoLagg.log(Level.WARNING, "Chunk read at " + chunk.toString() + " contains [" + realX + ", " + realZ + "] instead");
		if (!chunk.getWorld().isChunkLoaded(realX, realZ) && this.a(chunk.world, realX, realZ) == null) {
			//no chunk is present here, we can safely move this chunk
		    //get rid of the old chunk
			this.clearChunk(chunk.x, chunk.z);
			//write the chunk to the correct position
			DataOutputStream dataoutputstream = RegionFileCache.c(this.sourcefolder, realX, realZ);
	        NBTCompressedStreamTools.a(compound, (DataOutput) dataoutputstream);
	        dataoutputstream.close();
        
			NoLagg.log(Level.WARNING, "The chunk has been relocated to the correct location the region files.");
		} else {
			//a chunk is present at the correct location, alter the x/z location of the data instead
			NBTTagCompound level = compound.getCompound("Level");
			level.setInt("xPos", chunk.x);
			level.setInt("zPos", chunk.z);
			//rewrite the chunk
			DataOutputStream dataoutputstream = RegionFileCache.c(this.sourcefolder, chunk.x, chunk.z);
	        NBTCompressedStreamTools.a(compound, (DataOutput) dataoutputstream);
	        dataoutputstream.close();

			NoLagg.log(Level.WARNING, "The saved chunk coordinates have been altered to match the requested coordinates.");
		}
	}
		
	/*
	 * This is called when a chunk has to be loaded from file
	 */
	@Override
	public Chunk a(World world, int i, int j) {
		if (NoLagg.useBufferedChunkLoading) {
			try {
				BufferedChunk chunk = BufferedChunk.removeUnloaded(world, i, j);
				if (chunk != null) chunk.clearData();
				
				NBTTagCompound nbttagcompound = null;
				ChunkCoordIntPair chunkcoordintpair = new ChunkCoordIntPair(i, j);
				
				//try to load this chunk from the saving queue
				synchronized (this.lock) {
					if (this.pendingsave.contains(chunkcoordintpair)) {
						Object obj;
						for (int k = 0; k < this.savequeue.size(); ++k) {
							obj = this.savequeue.get(k);
							if (getPendingA(obj).equals(chunkcoordintpair)) {
								nbttagcompound = getPendingB(obj);
								if (chunk == null) chunk = new BufferedChunk(world, i, j);
								break;
							}
						}
					}
				}

				//try to load this chunk from the region file cache
				if (nbttagcompound == null) {
					DataInputStream datainputstream = RegionFileCache.b(this.sourcefolder, i, j);
					if (datainputstream == null) return null;
					
					chunk = new BufferedChunk(world, i, j);

					//custom memory-efficient loading system
					try {
						nbttagcompound = chunk.getBufferedTag(datainputstream);
					} catch (ZipException ex) {
						NoLagg.log(Level.SEVERE, "Chunk data of " + chunk.toString() + " got corrupted (not in GZIP format) and has been removed.");
						this.clearChunk(i, j);
						return null;
					} catch (Exception ex) {
						NoLagg.log(Level.SEVERE, "An exception occurred while reading chunk " + chunk.toString() + ":");
						ex.printStackTrace();
						NoLagg.log(Level.WARNING, "Native methods will now try to read it...");	
						
						//failure, Use native coding instead
						datainputstream = RegionFileCache.b(this.sourcefolder, i, j);
						if (datainputstream == null) return null;
						nbttagcompound = NBTCompressedStreamTools.a((DataInput) datainputstream);
					}
				}
				
				//simple check
				if (nbttagcompound == null) return null;
				
				//checks and load
				if (nbttagcompound.hasKey("Level")) {
					NBTTagCompound level = nbttagcompound.getCompound("Level");
					if (level.hasKey("Blocks")) {
						//actual chunk loading from NBT compound
						try {
							chunk.load(this, nbttagcompound, level);
							chunk.h();
							return chunk;
						} catch (WrongChunkCoordinatesException ex) {
							//try to load it again; maybe it got freed now?
							//otherwise returns null so the chunk provider can generate it
							return this.a(world, i, j);
						} catch (Exception ex) {
							NoLagg.log(Level.SEVERE, "An exception occurred while loading NBT data of chunk " + chunk.toString() + ":");
							ex.printStackTrace();
							NoLagg.log(Level.WARNING, "Native methods will now try to load it...");	
							
							//failure, use native coding instead
							Chunk c = ChunkLoader.a(world, level);
							c.h();
							return c;
						}
					} else {
						System.out.println("Chunk file at " + i + "," + j + " is missing block data, skipping");
						return null;
					}
				} else {
					System.out.println("Chunk file at " + i + "," + j + " is missing level data, skipping");
					return null;
				}
			} catch (Throwable t) {
				NoLagg.log(Level.SEVERE, "Buffered chunk loader faced an incompatibility issue with the current build:");
				t.printStackTrace();
				NoLagg.log(Level.INFO, "Buffered chunk loading has been disabled until a server reload or restart is performed.");
				NoLagg.useBufferedChunkLoading = false;
			}
		}
		return super.a(world, i, j);
	}

}
